"""
model_fasterwhisper.py
"""

from faster_whisper import WhisperModel

class FasterWhisperModel:
    """
    A singleton class for managing an instance of the FasterWhisper model.

    Attributes:
        model (WhisperModel): The FasterWhisper model instance.
    """
    _instance = None
    device:str = "auto"
    model:str = "large-v3"

    def __new__(cls):
        """
        Create a new instance of the class if it doesn't exist, otherwise, return the existing 
        instance.

        Returns:
            FasterWhisperModel: The FasterWhisperModel instance.
        """
        if not cls._instance:
            cls._instance = super(FasterWhisperModel, cls).__new__(cls)
            cls._instance._init_once(device=FasterWhisperModel.device, model=FasterWhisperModel.model)
        return cls._instance

    # pylint: disable=attribute-defined-outside-init
    def _init_once(self, model:str, device:str):
        """
        Initialize the FasterWhisperModel instance.
        """
        if device == "cpu":
            self.model = WhisperModel(model_size_or_path=model, device=device, compute_type="float32")
        elif device == "cuda":
            self.model = WhisperModel(model_size_or_path=model, device=device, compute_type="float16")
        else:
            self.model = WhisperModel(model_size_or_path=model)

    def unload(self):
        """
        Unload the model by setting the model attribute to None and clearing the instance.
        """
        self.model = None
        FasterWhisperModel._instance = None



from .detect_language import DetectLanguage
from .detect_language_whisper import DetectLanguageWhisper
from .detect_language_fasterwhisper import DetectLanguageFasterWhisper
from .detect_language_file import DetectLanguageFile



import logging
import numpy as np
from numpy import ndarray

from ..transcription import Transcription, Word, Utterance
from ..models.model_fasterwhisper import FasterWhisperModel
from .transcribe_speech import TranscribeSpeech


LOG = logging.getLogger(__name__)

class TranscribeSpeechFasterWhisper(TranscribeSpeech):
    """
    Implementation of TranscribeSpeech using FasterWhisper model.

    Attributes:
        None
    """

    def _get_utterance_from_segment(self, segment, info, speaker, speech_offset):
        """
        Extracts Utterance information from a segment.

        Args:
            segment: Segment information from FasterWhisper model.
            info: Additional information about the segment.
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.

        Returns:
            Utterance: Utterance object containing information about the transcribed segment.
        """

        start = speech_offset + segment.start
        end = speech_offset + segment.end
        silence_prob = segment.no_speech_prob
        utterance = Utterance(
            start=start,
            end=end,
            language=info.language,
            confidence=np.exp(segment.avg_logprob),
            speaker=speaker,
            words=[Word(text=w.word, start=speech_offset + w.start, end=speech_offset + w.end,
                        confidence=(1 - silence_prob) * w.probability) for w in segment.words]
        )
        return utterance

    def execute_segment(self, speech_segment_float32_16khz: ndarray,
                        speaker: str = "speaker", speech_offset: float = 0,
                        language: str = None, prompt: str = "",
                        beams: int = 9, **kwargs: dict) -> Transcription:
        """
        Executes transcription on a speech segment using FasterWhisper model.

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language (str): Target language for transcription.
            prompt (str): Optional transcription prompt.
            beams (int): Number of beams for decoding.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the transcribed information.
        """
        transcription = Transcription()
        if 'device' in kwargs:
            FasterWhisperModel.device = kwargs['device']
        if 'model_fasterwhisper' in kwargs:
            FasterWhisperModel.model = kwargs['model_fasterwhisper']
        model = FasterWhisperModel().model

        segments, info = model.transcribe(
            word_timestamps=True,
            audio=speech_segment_float32_16khz,
            initial_prompt=prompt,
            language=language,
            beam_size=beams,
            best_of=beams,
            patience=1,
            vad_filter=True
        )
        for segment in segments:
            utterance = self._get_utterance_from_segment(segment, info, speaker, speech_offset)
            transcription.append(utterance)
            LOG.info(f"[silent={segment.no_speech_prob:.2%};" +
                     f"temp={segment.temperature:.2f};" + 
                     f"comp={segment.compression_ratio:.2%}]" + 
                     f"{utterance.get_colour_text()}")
        return transcription

    def detect_language(self, speaker: str, speech_offset: float, speech_segment_float32_16khz: ndarray,
                        languages=None, **kwargs: dict) -> Transcription:
        """
        Detects language in a speech segment using FasterWhisper model.

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language_file (str): File path for saving language information.
            languages (list): List of target languages.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the detected language information.
        """
        transcription = Transcription()

        if 'device' in kwargs:
            FasterWhisperModel.device = kwargs['device']
        if 'model_fasterwhisper' in kwargs:
            FasterWhisperModel.model = kwargs['model_fasterwhisper']
        model = FasterWhisperModel().model

        features = model.feature_extractor(speech_segment_float32_16khz)

        segment = features[:, : model.feature_extractor.nb_max_frames]
        encoder_output = model.encode(segment)
        # results is a list of tuple[str, float] with language names and
        # probabilities.
        results = model.model.detect_language(encoder_output)[0]
        # Parse language names to strip out markers
        all_language_probs = [(token[2:-2], prob) for (token, prob) in results]
        detected_languages = {}
        for l, p in all_language_probs:
            detected_languages[l] = p

        best_language = None
        best_language_prob = 0
        if languages is not None and len(languages) > 0:
            for lang in languages:
                if lang in detected_languages:
                    if detected_languages[lang] > best_language_prob:
                        best_language_prob = detected_languages[lang]
                        best_language = lang

        if best_language is None:
            best_language = max(detected_languages, key=detected_languages.get)
            best_language_prob = detected_languages[best_language]

        utterance = Utterance(
            start=speech_offset,
            end=speech_offset + len(speech_segment_float32_16khz) / 16000,
            language=best_language,
            confidence=best_language_prob,
            speaker=speaker,
            words=[])
        transcription.append(utterance)
        LOG.info(utterance.get_colour_text())
        return transcription
    


import logging
import os
from pathlib import Path
import yaml


LOG = logging.getLogger(__name__)


class Context:
    source_file_path: str
    work_directory_path: str
    audio_file_path: str
    voice_file_path: str
    diarization_file: str
    language_file: str
    transcription_path: str
    output_file: str
    languages: [str]
    nb_speakers: int
    log_level: int
    device: str
    model_fasterwhisper:str
    model_whisper:str
    model_pyannote_segmentation:str
    model_pyannote_diarization:str
    model_mdx:str
    model_speechbrain_vad:str
    has_ffmpeg: bool
    beams:int
    transcribe_only:bool

    def __init__(self, source_file: str, out_dir: str = "out", languages: [str] = None,
                  nb_speakers=1, log_level=logging.WARNING, device:str="cuda",
                  model_fasterwhisper="large-v3", model_whisper="large-v3",
                  model_pyannote_segmentation:str = "pyannote/segmentation-3.0",
                  model_pyannote_diarization:str = "pyannote/speaker-diarization-3.1",
                  model_mdx:str = 'Kim_Vocal_2', model_speechbrain_vad:str = "speechbrain/vad-crdnn-libriparty"):
        super().__init__()
        self.source_file_path = os.path.abspath(source_file)
        self.work_directory_path = os.path.abspath(out_dir)
        source_name, _ = os.path.splitext(os.path.basename(self.source_file_path))
        self.audio_file_path = str(Path.joinpath(Path(self.work_directory_path), f"{source_name}.audio.wav"))
        self.voice_file_path = str(Path.joinpath(Path(self.work_directory_path), f"{source_name}.voices.wav"))
        self.diarization_file = str(Path.joinpath(Path(self.work_directory_path), f"{source_name}.rttm"))
        self.language_file = str(Path.joinpath(Path(self.work_directory_path), f"{source_name}.lang.json"))
        self.transcription_path = str(Path.joinpath(Path(self.work_directory_path), f"{source_name}.script.json"))
        self.output_file = str(Path.joinpath(Path(self.work_directory_path), f"{source_name}.output"))
        self.languages = languages
        self.min_speakers = nb_speakers
        self.max_speakers = nb_speakers
        self.log_level = log_level
        self.device = device
        self.model_fasterwhisper = model_fasterwhisper
        self.model_whisper = model_whisper
        self.model_pyannote_segmentation = model_pyannote_segmentation
        self.model_pyannote_diarization = model_pyannote_diarization

        self.model_mdx = model_mdx
        self.model_speechbrain_vad = model_speechbrain_vad
        self.has_ffmpeg = True
        self.transcribe_only = False
        self.beams = 9

    def __str__(self):
        return self.to_yaml()

    def to_yaml(self):
        return yaml.dump(self.to_dict())

    @staticmethod
    def from_yaml(source_file:str, out_dir:str, text:str):
        dict_data = yaml.safe_load(text)
        return Context.from_dict(source_file=source_file, out_dir=out_dir, dict_data=dict_data)

    @staticmethod
    def from_dict(source_file:str, out_dir:str, dict_data:dict):
        obj:Context = Context(source_file=source_file, out_dir=out_dir)
        obj.__dict__.update(dict_data)
        return obj

    def to_dict(self) -> dict:
        return self.__dict__



import os
import logging
from abc import abstractmethod
import numpy as np
from numpy import ndarray
from pyannote.core import Annotation, Segment
from pyannote.database.util import load_rttm

from ..transcription import Transcription
from ..speaker_diarization import DiarizeSpeakersSpeechBrain
from ..wav_conversion import ConvertToWav
from ..filter import Filter

LOG = logging.getLogger(__name__)


class TranscribeSpeech(Filter):
    """
    Abstract class for transcribing audio.

    Attributes:
        None
    """

    @abstractmethod
    def execute_segment(self, speech_segment_float32_16khz: ndarray,
                        speaker: str = "speaker", speech_offset: float = 0,
                        language: str = None, prompt: str = "",
                        **kwargs: dict) -> Transcription:
        """
        Abstract method for executing transcription on a speech segment.

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language (str): Target language for transcription.
            prompt (str): Optional transcription prompt.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the transcribed information.
        """


    def _optimize_sequence(self, sequence: dict, full_duration: float) -> dict:
        """
        Optimizes the sequence by ensuring language and speaker information is set consistently.

        Args:
            sequence (dict): Dictionary containing language and speaker information at different time points.
            full_duration (float): Full duration of the audio in seconds.

        Returns:
            dict: Optimized sequence.
        """
        # ensure that language is set from the start
        sequence = dict(sorted(sequence.items()))
        if int(0) not in sequence.keys():
            sequence[0] = sequence[sequence.keys()[0]]

        # ensure that language is set until the end
        last_index: int = int(full_duration * 16000)
        sequence = dict(sorted(sequence.items()))
        if last_index not in sequence:
            sequence[last_index] = sequence[next(iter(reversed(sequence.keys())))]

        # propagate language over short gaps
        changed2 = True
        while changed2:
            changed2 = False
            changed = True
            while changed:
                changed = False
                index_seq = sorted(sequence.items())
                _, first = index_seq[0]
                _, after = index_seq[1]
                if first['language'] == "none" and after['language'] != "none":
                    first['language'] = after['language']
                    changed = True
                if first['speaker'] == "none" and after['speaker'] != "none":
                    first['speaker'] = after['speaker']
                    changed = True

                _, last = index_seq[len(index_seq) - 1]
                _, prev = index_seq[len(index_seq) - 2]
                if last['language'] == "none" and prev['language'] != "none":
                    last['language'] = prev['language']
                    changed = True
                if last['speaker'] == "none" and prev['speaker'] != "none":
                    last['speaker'] = prev['speaker']
                    changed = True

                for i in range(1, len(index_seq) - 1):
                    _, prev = index_seq[i - 1]
                    _, after = index_seq[i + 1]
                    _, cur = index_seq[i]

                    if cur['language'] == "none":
                        if prev['language'] == after['language'] and cur['language'] != prev['language']:
                            cur['language'] = prev['language']
                            changed = True

                        elif prev['language'] != "none":
                            cur['language'] = prev['language']
                            changed = True

                        elif after['language'] != "none":
                            cur['language'] = after['language']
                            changed = True

                    if cur['speaker'] == "none":
                        if prev['speaker'] == after['speaker'] and cur['speaker'] != prev['speaker']:
                            cur['speaker'] = prev['speaker']
                            changed = True
                        elif prev['speaker'] != "none":
                            cur['speaker'] = prev['speaker']
                            changed = True

                        elif after['speaker'] != "none":
                            cur['speaker'] = after['speaker']
                            changed = True

                min_duration = 1
                current_start = 0
                current_state = None
                for time, state in sorted(sequence.items()):
                    if current_state is None:
                        current_start = time
                        current_state = state
                    else:
                        duration = (time - current_start) / 16000
                        if duration < min_duration:
                            if current_state['language'] == "none" and state['language'] != "none":
                                current_state['language'] = state['language']
                                changed = True
                                changed2 = True
                            if current_state['speaker'] == "none" and state['speaker'] != "none":
                                current_state['speaker'] = state['speaker']
                                changed = True
                                changed2 = True
                        current_start = time
                        current_state = state

            changed = True
            while changed:
                changed = False
                min_duration = 1
                current_start = 0
                current_state = None
                for time, state in reversed(sorted(sequence.items())):
                    if current_state is None:
                        current_start = time
                        current_state = state
                    else:
                        duration = -(time - current_start) / 16000
                        if duration < min_duration:
                            if current_state['language'] == "none" and state['language'] != "none":
                                current_state['language'] = state['language']
                                changed = True
                                changed2 = True
                            if current_state['speaker'] == "none" and state['speaker'] != "none":
                                current_state['speaker'] = state['speaker']
                                changed = True
                                changed2 = True
                        current_start = time
                        current_state = state

        return sequence

    def _prepare_second_pass_diarization(self, diarization: Annotation, detected_languages: Transcription) -> dict:
        """
        Prepares the diarization for the second pass by creating a sparse sequence.

        Args:
            diarization (Annotation): Speaker diarization information.
            detected_languages (Transcription): Transcription containing detected language information.

        Returns:
            dict: Sparse sequence for the second pass.
        """

        sparse_sequence = {}
        sparse_sequence[0] = {"speaker": "none", "language": "none"}
        for turn, _, speaker in diarization.itertracks(yield_label=True):
            sparse_sequence.setdefault(int(turn.start * 16000), {}).update(
                {"speaker": speaker})
            sparse_sequence.setdefault(int(turn.end * 16000), {}).update(
                {"speaker": "none"})

        for utterance in detected_languages.utterances:
            sparse_sequence.setdefault(int(utterance.start * 16000), {}).update(
                {"language": utterance.language})
            sparse_sequence.setdefault(int(utterance.end * 16000), {}).update(
                {"language": "none"})

        state = {"language": "none", "speaker": "none"}
        sequence = {0: state}
        for index, sparse_state in sorted(sparse_sequence.items()):
            state = dict(state)
            state.update(sparse_state)
            sequence[int(index)] = state

        return sequence

    @staticmethod
    def apply_cosine_fade(samples: ndarray, fade_start: int, fade_end: int, start: int, end: int):
        """
        Applies a cosine fade to a portion of audio.

        Args:
            samples (ndarray): Audio samples.
            fade_start (int): Start index of the fade.
            fade_end (int): End index of the fade.
            start (int): Start index of the audio portion.
            end (int): End index of the audio portion.
        """
        if fade_start > 0:
            fade_out = np.cos(np.linspace(0, np.pi / 2, fade_start))
            samples[start:start + fade_start] *= fade_out

        if fade_end > 0:
            fade_in = np.cos(np.linspace(np.pi / 2, 0, fade_end))
            samples[end - fade_end:end] *= fade_in

        # Mute the audio in between fade-in and fade-out
        samples[start + fade_start:end - fade_end] = 0.0

    def execute_for_speaker_and_language(self, speech_segment_float32_16khz, sequence,
                                         speaker: str, language: str, **kwargs: dict):
        """
        Executes transcription for a specific speaker and language.

        Args:
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            sequence (dict): Dictionary containing language and speaker information at different time points.
            speaker (str): The speaker identifier.
            language (str): Target language for transcription.

        Returns:
            Transcription: Transcription object containing the transcribed information.
        """
        audio_lang = speech_segment_float32_16khz.copy()

        mute_seq = {}
        mute_start = None
        mute_end = None
        for index, info in sequence.items():
            if info['speaker'] != speaker or (
                    (language is None or info['language'] != language) and info['language'] != "none"):
                if mute_start is None:
                    mute_start = index
                mute_end = index
            else:
                if mute_start is not None:
                    mute_end = index
                    mute_seq[mute_start] = mute_end
                    mute_start = None
                    mute_end = None
        if mute_start is not None:
            mute_seq[mute_start] = len(audio_lang)

        for mute_start, mute_end in mute_seq.items():
            fade_duration = 0.1 * 16000  # Adjust this parameter for the fade duration
            fade_start = min(len(audio_lang) - mute_start, int(max(0, min(mute_start, fade_duration / 2))) * 2)
            fade_end = int(max(0, min(len(audio_lang) - mute_end, fade_duration / 2))) * 2
            TranscribeSpeech.apply_cosine_fade(
                samples=audio_lang,
                fade_start=int(fade_start),
                fade_end=int(fade_end),
                start=int(mute_start - fade_start / 2),
                end=int(mute_end + fade_end / 2),
            )

        # speechbrain throws an exception when processing audio shorter than 30s
        # https://github.com/speechbrain/speechbrain/issues/2334
        audio_lang = DiarizeSpeakersSpeechBrain.pad_audio_to_duration(audio_lang, 31, 16000)

        work_directory_path = kwargs['work_directory_path'] or "."
        speaker_lang_audio_path = os.path.join(work_directory_path, f"{speaker}-{language}.wav")
        ConvertToWav.save_float32_16khz_mono_audio(audio_lang, speaker_lang_audio_path)

        segments = DiarizeSpeakersSpeechBrain().diarize_on_silences(
            audio_path=speaker_lang_audio_path, **kwargs)
        prompt = "Hello and welcome. This is my presentation."
        whole_transcription = Transcription()
        for turn, _, _ in segments.itertracks(yield_label=True):
            sample_start = int(turn.start * 16000)
            sample_end = int(turn.end * 16000)
            transcription: Transcription = self.execute_segment(
                speaker=speaker,
                speech_offset=turn.start,
                speech_segment_float32_16khz=audio_lang[sample_start:sample_end],
                prompt=prompt,
                language=language,
                **kwargs)
            prompt = transcription.get_text()
            for u in transcription.utterances:
                whole_transcription.append(u)
        return whole_transcription

    # pylint: disable=unused-argument
    # pylint: disable=arguments-differ
    def execute(self, voice_file_path:str, language_file:str,
                transcription_path: str, diarization_file:str, languages: list,
                **kwargs: dict) -> Transcription:
        """
        Executes the transcription process for multiple speakers and languages.

        Args:
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            detected_languages (Transcription): Transcription containing detected language information.
            transcription_path (str): File path to save the final transcription result.
            diarization (Annotation): Speaker diarization information.
            languages (list): List of target languages.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the final result.
        """

        speech_segment_float32_16khz = ConvertToWav.load_float32_16khz_mono_audio(
            input_file=voice_file_path, device=kwargs['device'])
        if language_file is None:
            detected_languages: Transcription = Transcription()
        else:
            detected_languages: Transcription = Transcription.load(language_file)
        if diarization_file is None:
            rttms = None
            diarization = Annotation(uri="waveform")
            diarization[Segment(0, len(speech_segment_float32_16khz)/16000)] = "speaker"
        else:
            rttms = load_rttm(diarization_file)
            diarization = next(iter(rttms.values()))

        full_transcription = Transcription()
        sequence = self._prepare_second_pass_diarization(diarization=diarization, detected_languages=detected_languages)
        sequence = self._optimize_sequence(sequence=sequence, full_duration=len(speech_segment_float32_16khz) / 16000)

        speakers = set()
        for info in sequence.values():
            speakers.add(info['speaker'])

        if languages is None or len(languages) == 0:
            for speaker in speakers:
                transcription: Transcription = self.execute_for_speaker_and_language(
                    speech_segment_float32_16khz=speech_segment_float32_16khz, sequence=sequence, speaker=speaker,
                    language=None,
                    **kwargs)
                for utterance in transcription.utterances:
                    full_transcription.append(utterance)
        else:
            for language in languages:
                for speaker in speakers:
                    transcription: Transcription = self.execute_for_speaker_and_language(
                        speech_segment_float32_16khz=speech_segment_float32_16khz, sequence=sequence, speaker=speaker,
                        language=language,
                        **kwargs)
                    for utterance in transcription.utterances:
                        full_transcription.append(utterance)

        full_transcription = full_transcription.regroup_by_words()
        full_transcription.save(transcription_path)
        LOG.info(full_transcription.get_colour_text())
        return full_transcription



from numpy import ndarray

from .isolate_voices import IsolateVoices
from ..wav_conversion import ConvertToWav


class IsolateVoicesFile(IsolateVoices):
    """
    Voice isolation using an existing audio file.

    This class inherits from IsolateVoices and implements the voice isolation process using an existing audio file.
    It loads the audio file, returning its waveform as a NumPy array.

    Attributes:
        None
    """

    def execute(self, audio_file_path: str, voice_file_path: str, **kwargs) -> ndarray:
        """
        Execute the voice isolation process using an existing audio file.

        Args:
            audio_file_path (str): Path to the existing audio file for voice isolation (not used in this method)..
            voice_file_path (str): Path to save the isolated voice audio file.
            **kwargs: Additional parameters (not used in this method).

        Returns:
            ndarray: NumPy array representing the isolated voice audio.
        """
        # Load the existing audio file as a NumPy array
        waveform: ndarray = ConvertToWav.load_float32_16khz_mono_audio(voice_file_path, kwargs['device'])

        return waveform



from pyannote.core import Annotation
from pyannote.database.util import load_rttm
from .diarize_speakers import DiarizeSpeakers


class DiarizeSpeakersFile(DiarizeSpeakers):
    """
    Diarization implementation that loads speaker information from an existing RTTM file.

    This class inherits from DiarizeSpeakers and implements diarization by loading speaker 
    information from a pre-existing RTTM (Rich Transcription Time Marked) file.

    Attributes:
        None
    """

    def execute(self, voice_file_path: str, diarization_file: str, min_speakers: int = 1, max_speakers: int = None,
                **kwargs: dict) -> Annotation:
        """
        Execute the diarization process by loading speaker information from an RTTM file.

        Args:
            voice_file_path (str): Path to the input audio file (not used in this method).
            diarization_file (str): Path to the RTTM file containing speaker information.
            min_speakers (int, optional): Minimum number of expected speakers. Default is 1.
            max_speakers (int, optional): Maximum number of expected speakers. Default is None (unbounded).
            TOKEN_HUGGINGFACE (str, optional): Hugging Face token for customization. Default is None.
            **kwargs (dict): Additional parameters (not used in this method).

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        # Load speaker information from the provided RTTM file
        rttms = load_rttm(diarization_file)

        # Return the first speaker information available
        return next(iter(rttms.values()))



import logging
from pyannote.core import Annotation, Segment
from pydub import AudioSegment

from .diarize_speakers import DiarizeSpeakers

LOG = logging.getLogger(__name__)


class DiarizeSpeakersNone(DiarizeSpeakers):
    """
    Diarization implementation that considers the entire audio as a single diarization segment.

    This class inherits from DiarizeSpeakers and implements diarization by considering the entire audio
    as a single diarization segment, assigned to a default speaker label.

    Attributes:
        None
    """

    def execute(self, voice_file_path: str, diarization_file: str, min_speakers: int = 1, max_speakers: int = None,
                **kwargs: dict) -> Annotation:
        """
        Execute the diarization process by considering the entire audio as a single diarization segment.

        Args:
            voice_file_path (str): Path to the input audio file.
            diarization_file (str): Path to the output RTTM (Rich Transcription Time Marked) file.
            min_speakers (int, optional): Minimum number of expected speakers. Default is 1.
            max_speakers (int, optional): Maximum number of expected speakers. Default is None (unbounded).
            **kwargs (dict): Additional parameters (not used in this method).

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        # Initialize an empty diarization annotation
        diarization: Annotation = Annotation()

        # Load the audio as a Pydub AudioSegment
        sound: AudioSegment = AudioSegment.from_file(voice_file_path)

        # Create a diarization segment covering the entire audio with a default speaker label
        diarization[Segment(0, sound.duration_seconds)] = "Speaker"

        # Write the diarization to the provided RTTM file
        with open(diarization_file, "w", encoding="utf-8") as f:
            diarization.write_rttm(f)

        # Log the diarization for information
        LOG.info(diarization)

        return diarization



import logging
from numpy import ndarray

from ..transcription import Transcription
from .detect_language import DetectLanguage

LOG = logging.getLogger(__name__)


class DetectLanguageFile(DetectLanguage):
    """
    Class for loading language detection results from a file.

    Attributes:
        None
    """

    def execute_segment(self, speaker: str, speech_offset: float, speech_segment_float32_16khz: ndarray,
                        languages=None, **kwargs: dict) -> Transcription:
        """
        Placeholder method for executing language detection on a speech segment.

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            languages (Optional[List[str]]): List of target languages for detection.
            **kwargs (dict): Additional keyword arguments for customization.

        Raises:
            NotImplementedError: This method is not implemented for file-based language detection.

        Returns:
            Transcription: Placeholder return value.
        """
        raise NotImplementedError

    # pylint: disable=arguments-differ
    def execute(self, language_file: str, **kwargs: dict) -> Transcription:
        """
        Loads language detection results from a file.

        Args:
            diarization (Annotation): Speaker diarization information (unused in this case).
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz 
                (unused in this case).
            language_file (str): File path to load the detected language information.
            languages (Optional[List[str]]): List of target languages (unused in this case).
            **kwargs (dict): Additional keyword arguments (unused in this case).

        Returns:
            Transcription: Transcription object containing the loaded language information.
        """
        transcription = Transcription.load(language_file)
        return transcription



from abc import ABC, abstractmethod

# pylint: disable=unused-argument
class Filter(ABC):
    @abstractmethod
    def execute(self, **kwargs: dict):
        ...
    def load(self, **kwargs: dict):
        ...

    def unload(self, **kwargs: dict):
        ...
        


from .write_transcript import WriteTranscript
from .write_transcript_docx import WriteTranscriptDocx
from .write_transcript_ass import WriteTranscriptAss
from .write_transcript_stdout import WriteTranscriptStdout



import logging
import os
import torchaudio
import torch

from pyannote.core import Annotation, Segment
from pyannote.audio import Model
from pyannote.audio.pipelines import VoiceActivityDetection
from pyannote.audio import Pipeline
from pyannote.audio.pipelines.utils.hook import ProgressHook

from ..wav_conversion import ConvertToWav
from .diarize_speakers import DiarizeSpeakers

LOG = logging.getLogger(__name__)


class DiarizeSpeakersPyannote(DiarizeSpeakers):
    """
    Diarization implementation using PyAnnote for speaker segmentation.

    This class inherits from DiarizeSpeakers and implements diarization using PyAnnote for speaker segmentation.

    Attributes:
        None
    """

    def diarize_on_silences(self, voice_file_path: str, model_pyannote_segmentation:str,
                            huggingface_token: str, **kwargs: dict) -> Annotation:
        """
        Diarize speakers based on silences using PyAnnote.

        Args:
            audio_file (str): Path to the input audio file.
            huggingface_token (str): Hugging Face token for model access.

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        LOG.info(f"Loading model {model_pyannote_segmentation}")
        model = Model.from_pretrained(model_pyannote_segmentation, use_auth_token=huggingface_token)
        if model is None:
            LOG.error(f"Failed to retrieve model {model_pyannote_segmentation}")
            raise FileNotFoundError

        pipeline = VoiceActivityDetection(segmentation=model)
        pipeline.to(torch.device(kwargs['device']))

        hyper_parameters = {
            # remove speech regions shorter than that many seconds.
            "min_duration_on": 0.0,
            # fill non-speech regions shorter than that many seconds.
            "min_duration_off": 0.5
        }
        pipeline.instantiate(hyper_parameters)
        vad: Annotation = pipeline(voice_file_path)
        vad.uri = "waveform"

        LOG.info(f"Unloading model {model_pyannote_segmentation}")
        del model
        return vad

    @staticmethod
    def _load_audio(audio_path: str) -> tuple:
        """
        Loads the audio waveform and sample rate from an audio file.

        Args:
            audio_path (str): Path to the input audio file.

        Returns:
            tuple: Tuple containing waveform and sample rate.
        """
        waveform, sample_rate = torchaudio.load(audio_path)

        resampler = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=16000)
        waveform = resampler(waveform)
        return waveform, 16000

    @staticmethod
    def _get_device(**kwargs:dict) -> torch.device:
        """
        Retrieves the device (CPU or CUDA) to use for diarization.

        Returns:
            torch.device: Device to use for diarization.
        """
        device = torch.device(kwargs['device'])
        LOG.info(f"Using device: {device.type}")
        return device

    @staticmethod
    def _load_pipeline(device: torch.device, model_pyannote_diarization:str, huggingface_token: str) -> Pipeline:
        """
        Loads the diarization pipeline.

        Args:
            device (torch.device): Device to use for diarization.
            huggingface_token (str): Hugging Face token for model access.

        Returns:
            Pipeline: PyAnnote diarization pipeline.
        """
        pipeline = Pipeline.from_pretrained(
            model_pyannote_diarization,
            use_auth_token=huggingface_token
        ).to(device)
        return pipeline

    def diarize_on_speakers(self, audio_file: str, num_speakers: int,
                             huggingface_token: str, model_pyannote_diarization:str, **kwargs:dict) -> Annotation:
        """
        Diarize speakers based on speaker count using PyAnnote.

        Args:
            audio_file (str): Path to the input audio file.
            num_speakers (int): Number of speakers to consider.
            huggingface_token (str): Hugging Face token for model access.

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        device: torch.device = DiarizeSpeakersPyannote._get_device(**kwargs)
        pipeline: Pipeline = DiarizeSpeakersPyannote._load_pipeline(
            device=device,
            model_pyannote_diarization=model_pyannote_diarization,
            huggingface_token=huggingface_token)
        waveform, sample_rate = DiarizeSpeakersPyannote._load_audio(audio_file)
        with ProgressHook() as hook:
            diarization: Annotation = pipeline({
                "waveform": waveform,
                "sample_rate": sample_rate,
                "window": "whole",
                "min_duration_on": 1.0,
                "min_duration_off": 0.0,
            },
                hook=hook,
                num_speakers=num_speakers,
                min_speakers=num_speakers,
                max_speakers=num_speakers
            )

        return diarization
    def execute(self, voice_file_path: str, diarization_file: str, min_speakers: int = 1, max_speakers: int = None,
                TOKEN_HUGGINGFACE: str = None, **kwargs: dict) -> Annotation:
        """
        Execute the diarization process using PyAnnote.

        Args:
            voice_file_path (str): Path to the input audio file.
            diarization_file (str): Path to the output RTTM (Rich Transcription Time Marked) file.
            min_speakers (int, optional): Minimum number of expected speakers. Default is 1.
            max_speakers (int, optional): Maximum number of expected speakers. Default is None (unbounded).
            TOKEN_HUGGINGFACE (str, optional): Hugging Face token for customization. Default is None.
            **kwargs (dict): Additional parameters (not used in this method).

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        diarization: Annotation = None
        huggingface_token: str = TOKEN_HUGGINGFACE or os.environ.get('TOKEN_HUGGINGFACE')
        if huggingface_token is None:
            LOG.warning("No HuggingFace token was provided (TOKEN_HUGGINGFACE)")
        if min_speakers == 1 and max_speakers == 1:
            try:
                diarization = self.diarize_on_silences(voice_file_path=voice_file_path,
                                                    huggingface_token=huggingface_token, **kwargs)
            except FileNotFoundError:
                # could not log model, default on full length diarization
                LOG.warning("Failed to compute diarization, defaulting on full length audio")
                diarization  = Annotation("waveform")
                audio = ConvertToWav.load_float32_16khz_mono_audio(input_file=voice_file_path, device=kwargs['device'])
                diarization[Segment(0, len(audio)/16000)] = "speaker"
        else:
            diarization = self.diarize_on_speakers(voice_file_path, max_speakers, huggingface_token, **kwargs)

        with open(diarization_file, "w", encoding="utf-8") as f:
            diarization.write_rttm(f)
        LOG.info(diarization)
        return diarization



from .isolate_voices import IsolateVoices
from .isolate_voices_demucs import IsolateVoicesDemucs
from .isolate_voices_mdx import IsolateVoicesMDX
from .isolate_voices_none import IsolateVoicesNone
from .isolate_voices_file import IsolateVoicesFile



import logging
import numpy as np
from numpy import ndarray

from ..transcription import Transcription, Word, Utterance
from ..models.model_whisper import WhisperModel
from .transcribe_speech import TranscribeSpeech

LOG = logging.getLogger(__name__)

class TranscribeSpeechWhisper(TranscribeSpeech):
    """
    Implementation of TranscribeSpeech using the OpenAI Whisper ASR model.

    Attributes:
        None
    """

    def _get_utterance_from_segment(self, seg, language, speaker, speech_offset):
        """
        Create an Utterance object from a Whisper ASR model segment.

        Args:
            seg (dict): Whisper ASR model segment.
            language (str): Detected language.
            speaker (str): Speaker identifier.
            speech_offset (float): Offset in seconds from the beginning of the audio.

        Returns:
            Utterance: Utterance object created from the segment.
        """
        start = seg['start']
        end = seg['end']
        avg_logprob = seg['avg_logprob']
        no_speech_prob = seg['no_speech_prob']
        utterance = Utterance(
            start=speech_offset + start,
            end=speech_offset + end,
            language=language,
            confidence=np.exp(avg_logprob),
            speaker=speaker,
            words=[Word(text=word['word'], start=speech_offset + word['start'], end=speech_offset + word['end'],
                        confidence=(1 - no_speech_prob) * word['probability']) for word in seg['words']]
        )
        return utterance

    def execute_segment(self, speech_segment_float32_16khz: ndarray,
                        speaker: str = "speaker", speech_offset: float = 0,
                        language: str = None, prompt: str = "",
                        beams: int = 9, **kwargs: dict) -> Transcription:
        """
        Execute transcription on a speech segment using the OpenAI Whisper ASR model.

        Args:
            speaker (str): Speaker identifier.
            speech_offset (float): Offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language (str): Target language for transcription.
            prompt (str): Optional transcription prompt.
            beams (int): Number of beams for transcription.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription result for the given speech segment.
        """
        transcription = Transcription()

        if 'device' in kwargs:
            WhisperModel.device = kwargs['device']
        if 'model_whisper' in kwargs:
            WhisperModel.model = kwargs['model_whisper']
        model = WhisperModel().model

        verbose = (kwargs['log_level'] or logging.WARNING) >= logging.INFO
        whisper_transcription = model.transcribe(
            word_timestamps=True,
            audio=speech_segment_float32_16khz,
            initial_prompt=prompt,
            language=language,
            beam_size=beams,
            best_of=beams,
            verbose=verbose
        )
        language = whisper_transcription['language']
        for segment in whisper_transcription['segments']:
            utterance = self._get_utterance_from_segment(segment, language, speaker, speech_offset)
            transcription.append(utterance)
            LOG.info(f"silent={segment['no_speech_prob']:.2%}|{utterance.get_colour_text()}")
        return transcription



import os
from numpy import ndarray
import demucs.separate

from ..wav_conversion import ConvertToWav
from .isolate_voices import IsolateVoices


class IsolateVoicesDemucs(IsolateVoices):
    """
    Voice isolation using the Demucs algorithm.

    This class inherits from IsolateVoices and implements the voice isolation process using the Demucs algorithm.

    Attributes:
        None
    """

    def execute(self, audio_file_path: str, voice_file_path: str, **kwargs) -> ndarray:
        """
        Execute the voice isolation process using Demucs.

        Args:
            audio_file_path (str): Path to the source audio file.
            voice_file_path (str): Path to save the isolated voice audio file.
            **kwargs: Additional parameters (not used in this method).

        Returns:
            ndarray: NumPy array representing the isolated voice audio.
        """
        # Use Demucs to separate vocals from the source audio
        demucs.separate.main([audio_file_path, "--two-stems", "vocals", "--name", "htdemucs_ft"])

        # Build the path to the separated vocals file
        filename_noext = os.path.splitext(os.path.basename(audio_file_path))[0]
        source_path_separated = f"separated/htdemucs_ft/{filename_noext}/vocals.wav"

        # Load the separated vocals as a NumPy array
        waveform: ndarray = ConvertToWav.load_float32_16khz_mono_audio(source_path_separated, kwargs['device'])

        # Save the isolated vocals to the destination path
        ConvertToWav.save_float32_16khz_mono_audio(waveform, voice_file_path)

        return waveform



import json
from typing import Optional
from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, DataClassJsonMixin



def red(text):
    return f"\033[91m{text}\033[00m"


def green(text):
    return f"\033[92m{text}\033[00m"


def yellow(text):
    return f"\033[93m{text}\033[00m"


def light_purple(text):
    return f"\033[94m{text}\033[00m"


def purple(text):
    return f"\033[95m{text}\033[00m"


def cyan(text):
    return f"\033[96m{text}\033[00m"


def light_gray(text):
    return f"\033[97m{text}\033[00m"


def black(text):
    return f"\033[0m{text}\033[00m"


def colorize(text, confidence):
    if confidence > 0.8:
        return green(text)
    elif confidence < 0.2:
        return light_purple(text)
    elif confidence < 0.4:
        return red(text)
    elif confidence < 0.6:
        return yellow(text)
    return text


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(eq=False, unsafe_hash=True)
class Word(DataClassJsonMixin):
    start: float
    end: float
    confidence: float
    text: str

    def __init__(self, text: str, start: float, end: float, confidence: float):
        self.start: float = start
        self.end: float = end
        self.confidence: float = confidence
        self.text: str = text

    def __json__(self):
        return self.__dict__

    def get_colour_text(self):
        return colorize(self.text, self.confidence)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(eq=False, unsafe_hash=True)
class Utterance(DataClassJsonMixin):
    words: list[Word]
    language: str
    speaker: str
    start: float
    end: float
    confidence: float
    silence_prob: float

    def __init__(self, speaker: str, words: [Word], language: str, start: float = None, end: float = None,
                 confidence: float = None, silence_prob: float = None):
        self.words: [Word] = words
        self.language: str = language
        self.speaker: str = speaker
        self.start: float = start or min((w.start for w in words), default=start)
        self.end: float = end or max((w.end for w in words), default=end)
        self.confidence: float = confidence or min((w.confidence for w in words), default=confidence)
        self.silence_prob: float = silence_prob

    def __str__(self):
        return ''.join(w.text for w in self.words)

    def __json__(self):
        return self.__dict__

    def get_colour_text(self):
        transcription = ''.join(word.get_colour_text() for word in self.words)
        return (f"[{self.start:.2f}->{self.end:.2f} {self.end - self.start:.2f}]" +
                f"[{self.language}][{self.confidence:.2%}] {self.speaker}: {transcription}")

    def get_text(self):
        return ''.join(word.text for word in self.words)

    def append(self, word: Word):
        self.words.append(word)
        self.start: float = min(self.start, word.start)
        self.end: float = max(self.end, word.end)
        self.confidence: float = min(self.confidence, word.confidence)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(eq=False, unsafe_hash=True)
class Transcription(DataClassJsonMixin):
    start: float
    end: float
    confidence: float
    utterances: list[Utterance]

    def __init__(self, start: float = None, end: float = None, confidence: float = None,
                 utterances: list[Utterance] = None):
        self.start = start
        self.end = end
        self.confidence = confidence
        self.utterances = utterances or []

    def append(self, utterance):
        self.utterances.append(utterance)
        self.utterances = sorted(self.utterances, key=lambda u: u.start)
        self.start = min(self.start or utterance.start, utterance.start)
        self.end = max(self.end or utterance.end, utterance.end)
        self.confidence = min(self.confidence or utterance.confidence, utterance.confidence)

    @staticmethod
    def load(json_file_path: str):
        with open(json_file_path, "r", encoding="utf-8") as f:
            parsed = json.load(f)
            return Transcription.from_dict(parsed)

    def save(self, json_file_path: str):
        with open(json_file_path, "w", encoding="utf-8") as f:
            json.dump(obj=self.to_dict(), fp=f, skipkeys=True, indent=2)

    def get_colour_text(self):
        return '\n'.join(utterance.get_colour_text() for utterance in self.utterances)

    def get_text(self):
        return '\n'.join(utterance.get_text() for utterance in self.utterances)

    def regroup_by_words(self):
        uttered_words: [Utterance] = []
        for utterance in self.utterances:
            for word in utterance.words:
                uttered_words.append(
                    Utterance(utterance.speaker, words=[word], language=utterance.language, start=word.start,
                              end=word.end, confidence=word.confidence, silence_prob=utterance.silence_prob))

        uttered_words = sorted(uttered_words, key=lambda u: u.start)

        grouped: Transcription = Transcription()
        state: Optional[Utterance] = None
        for uttered_word in uttered_words:
            if state is not None:
                if (state.language == uttered_word.language
                    and state.speaker == uttered_word.speaker
                    and uttered_word.start - state.end < 0.75):
                    state.append(uttered_word.words[0])
                else:
                    grouped.append(state)
                    state = uttered_word
            else:
                state = uttered_word
        if state is not None:
            grouped.append(state)
        return grouped



"""
__init__.py
"""
import warnings

# see https://github.com/pyannote/pyannote-audio/issues/1576
warnings.filterwarnings("ignore", category=UserWarning, module=r"pyannote\.audio\.core\.io")

# see https://github.com/asteroid-team/torch-audiomentations/issues/172
warnings.filterwarnings("ignore", category=UserWarning, module=r"torch_audiomentations\.utils\.io")

# see https://github.com/asteroid-team/torch-audiomentations/issues/172
warnings.filterwarnings("ignore", category=UserWarning, module=r".*")

# pylint: disable=wrong-import-position
from .context import Context
from .pipeline import Pipeline

__version__ = "0.1.6"



from abc import abstractmethod
from numpy import ndarray
from pyannote.core import Annotation
from pyannote.database.util import load_rttm
from ..wav_conversion import ConvertToWav

from ..transcription import Transcription, Utterance
from ..filter import Filter

class DetectLanguage(Filter):
    """
    Abstract base class for language detection.

    Attributes:
        None
    """

    @abstractmethod
    def execute_segment(self, speaker: str, speech_offset: float, speech_segment_float32_16khz: ndarray,
                        languages=None, **kwargs: dict) -> Transcription:
        """
        Abstract method to be implemented by subclasses for executing language detection on a speech segment.

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            languages (Optional[List[str]]): List of target languages for detection.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the detected language information.
        """

    @staticmethod
    def get_speech_segment(speech_segment_float32_16khz: ndarray, start: float, end: float) -> ndarray:
        """
        Extracts a subsegment from the given speech segment based on start and end time.

        Args:
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            start (float): Start time of the subsegment in seconds.
            end (float): End time of the subsegment in seconds.

        Returns:
            ndarray: Subsegment of the speech segment.
        """
        start_index = int(start * 16000)
        end_index = int(end * 16000)
        return speech_segment_float32_16khz[start_index:end_index]

    def identify_diarization_silences(self,
                                      speech_segment_float32_16khz: ndarray,
                                      diarization:Annotation,
                                      languages:[str],
                                      **kwargs:dict) -> Transcription:
        last_time: float = 0
        transcription = Transcription()

        for turn, _, speaker in diarization.itertracks(yield_label=True):
            skip_time = turn.start - last_time
            if skip_time > 0.25:
                segment_transcription: Transcription = self.execute_segment(
                    speaker="none",
                    speech_offset=last_time,
                    speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz,
                                                                                   last_time, turn.start),
                    languages=languages,
                    **kwargs)
                for utterance in segment_transcription.utterances:
                    transcription.append(utterance)
            last_time = turn.end

            segment_transcription: Transcription = self.execute_segment(
                speaker=speaker,
                speech_offset=turn.start,
                speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz, turn.start,
                                                                               turn.end),
                languages=languages,
                **kwargs)

            for utterance in segment_transcription.utterances:
                transcription.append(utterance)

        audio_end = len(speech_segment_float32_16khz) / 16000
        skip_time = audio_end - last_time
        if skip_time > 0.25:
            segment_transcription: Transcription = self.execute_segment(
                speaker="none",
                speech_offset=last_time,
                speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz, last_time,
                                                                               audio_end),
                languages=languages,
                **kwargs)
            for utterance in segment_transcription.utterances:
                transcription.append(utterance)

        return transcription

    def get_utterances_as_triplets(self, transcription:Transcription) -> [[Utterance]]:
        triplets: list[list[Utterance]] = []
        utterance = None
        for utterance in transcription.utterances:
            if len(triplets) == 0:
                triplets.append([utterance])
            if len(triplets) == 1:
                triplets[0].append(utterance)
                triplets.append([utterance])
            triplets[len(triplets) - 2].append(utterance)
            triplets[len(triplets) - 1].append(utterance)
            triplets.append([utterance])
        triplets[len(triplets) - 2].append(utterance)
        triplets[len(triplets) - 1].append(utterance)
        triplets[len(triplets) - 1].append(utterance)
        return triplets

    def detect_unknown_languages(self,
                                 speech_segment_float32_16khz:ndarray,
                                 triplets:[[Utterance]],
                                 languages:[str],
                                 **kwargs:dict):
        for triplet in triplets:
            center = triplet[1]

            if center.confidence < 0.8 or center.end - center.start < 1:
                votes = {}
                votes_confidence = {}
                for pad in [0.5, 1.0, 1.5, 2.0, 2.5]:
                    start = max(0, center.start - pad)
                    end = min(len(speech_segment_float32_16khz) / 16000, center.end + pad)

                    segment_transcription: Transcription = self.execute_segment(
                        speaker=center.speaker,
                        speech_offset=start,
                        speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz,
                                                                                       start, end),
                        languages=languages,
                        **kwargs)
                    confidence = 1
                    lang = set()
                    for utterance in segment_transcription.utterances:
                        if utterance.end >= center.start and utterance.start < center.end:
                            lang.add(utterance.language)
                            confidence = min(confidence, utterance.confidence)
                    if len(lang) == 1:
                        votes.setdefault(next(iter(lang)), 0)
                        votes[next(iter(lang))] = votes[next(iter(lang))] + 1
                        votes_confidence.setdefault(next(iter(lang)), confidence)
                        votes_confidence[next(iter(lang))] = max(confidence, votes_confidence[next(iter(lang))])
                max_votes = max(votes.values())
                max_confidence = max(v for k, v in votes_confidence.items() if votes[k] == max_votes)
                max_lang = next((k for k, v in votes_confidence.items() if v == max_confidence), None)
                if max_confidence > center.confidence:
                    if max_lang is not None and max_lang != center.language:
                        center.language = max_lang
                    center.confidence = max_confidence

    def fill_language_gaps(self,
                           triplets:[[Utterance]],
                           speech_segment_float32_16khz:ndarray,
                           languages:[str],
                           **kwargs:dict):
        changed = True

        i = 0
        while changed:
            changed = False
            i += 1
            for triplet in triplets:
                left = triplet[0]
                center = triplet[1]
                right = triplet[2]

                if center.confidence < 0.5 or center.end - center.start < 2:
                    if left.language == right.language and left.language != center.language:
                        start = left.start
                        end = right.end

                        segment_transcription: Transcription = self.execute_segment(
                            speaker=center.speaker,
                            speech_offset=start,
                            speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz,
                                                                                           start, end),
                            languages=languages,
                            **kwargs)
                        confidence = 1
                        lang = set()
                        for utterance in segment_transcription.utterances:
                            if utterance.end >= center.start and utterance.start < center.end:
                                lang.add(utterance.language)
                                confidence = min(confidence, utterance.confidence)
                        if confidence > center.confidence:
                            if len(lang) == 1 and next(iter(lang)) != center.language:
                                center.language = next(iter(lang))
                                changed = True
                            center.confidence = confidence

                    if left.language != center.language:
                        start = left.start
                        end = center.end

                        segment_transcription: Transcription = self.execute_segment(
                            speaker=center.speaker,
                            speech_offset=start,
                            speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz,
                                                                                           start, end),
                            languages=languages,
                            **kwargs)
                        confidence = 1
                        lang = set()
                        for utterance in segment_transcription.utterances:
                            if utterance.end >= center.start and utterance.start < center.end:
                                lang.add(utterance.language)
                                confidence = min(confidence, utterance.confidence)
                        if confidence > center.confidence:
                            if len(lang) == 1 and next(iter(lang)) != center.language:
                                center.language = next(iter(lang))
                                changed = True
                            center.confidence = confidence

                    if right.language != center.language:
                        start = center.start
                        end = right.end

                        segment_transcription: Transcription = self.execute_segment(
                            speaker=center.speaker,
                            speech_offset=start,
                            speech_segment_float32_16khz=DetectLanguage.get_speech_segment(speech_segment_float32_16khz,
                                                                                           start, end),
                            languages=languages,
                            **kwargs)
                        confidence = 1
                        lang = set()
                        for utterance in segment_transcription.utterances:
                            if utterance.end >= center.start and utterance.start < center.end:
                                lang.add(utterance.language)
                                confidence = min(confidence, utterance.confidence)
                        if confidence > center.confidence:
                            if len(lang) == 1 and next(iter(lang)) != center.language:
                                center.language = next(iter(lang))
                                changed = True
                            center.confidence = confidence

    # pylint: disable=arguments-differ
    def execute(self, diarization_file: str, voice_file_path:str,
                language_file: str, languages=None, **kwargs: dict) -> Transcription:
        """
        Executes language detection on the entire audio based on speaker diarization.

        Args:
            diarization (Annotation): Speaker diarization information.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language_file (str): File path to save the detected language information.
            languages (Optional[List[str]]): List of target languages for detection.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the detected language information.
        """

        rttms = load_rttm(diarization_file)
        diarization = next(iter(rttms.values()))

        if languages is not None and len(languages) == 1:
            transcription = Transcription()
            for turn, _, speaker in diarization.itertracks(yield_label=True):
                transcription.append(
                    Utterance(speaker=speaker, words=[],
                              language=languages[0],
                              start=turn.start, end=turn.end,
                              confidence=1.0, silence_prob=0.0))
        else:
            speech_segment_float32_16khz = ConvertToWav.load_float32_16khz_mono_audio(
                input_file=voice_file_path, device=kwargs['device'])

            transcription = self.identify_diarization_silences(
                speech_segment_float32_16khz=speech_segment_float32_16khz,
                diarization=diarization,
                languages=languages,
                **kwargs)

            # as a second pass, identify short segments with low confidence, and attempt to
            # merge with high confidence neighbors

            triplets: [[Utterance]] = self.get_utterances_as_triplets(transcription)

            self.detect_unknown_languages(
                speech_segment_float32_16khz=speech_segment_float32_16khz,
                triplets=triplets,
                languages=languages,
                **kwargs)

            self.fill_language_gaps(
                speech_segment_float32_16khz=speech_segment_float32_16khz,
                triplets=triplets,
                languages=languages,
                **kwargs)

        transcription.save(language_file)
        return transcription



import shutil
import logging
from numpy import ndarray
from audio_separator.separator import Separator

from ..wav_conversion import ConvertToWav
from .isolate_voices import IsolateVoices

class IsolateVoicesMDX(IsolateVoices):
    """
    Voice isolation using the MDX algorithm.

    This class inherits from IsolateVoices and implements the voice isolation process using the MDX algorithm.

    Attributes:
        None
    """

    # pylint: disable=arguments-differ
    def execute(self, audio_file_path: str, model_mdx:str, voice_file_path: str, **kwargs) -> ndarray:
        """
        Execute the voice isolation process using the MDX algorithm.

        Args:
            audio_file_path (str): Path to the source audio file.
            voice_file_path (str): Path to save the isolated voice audio file.
            **kwargs: Additional parameters (not used in this method).

        Returns:
            ndarray: NumPy array representing the isolated voice audio.
        """
        # Initialize the MDX separator
        separator = Separator(log_level=kwargs['log_level'] or logging.WARNING)
        separator.load_model(model_mdx)

        # Use MDX to separate vocals from the source audio
        output_file_paths = separator.separate(audio_file_path)

        # Move the generated files to the output directory
        instrument_audio = output_file_paths[0]
        voice_audio = output_file_paths[1]
        shutil.move(instrument_audio, f"{voice_file_path}-noise.wav")
        shutil.move(voice_audio, f"{voice_file_path}-voice.wav")

        # Load the separated vocals as a NumPy array
        waveform: ndarray = ConvertToWav.load_float32_16khz_mono_audio(
            f"{voice_file_path}-voice.wav", kwargs['device'])

        # Save the isolated vocals to the destination path
        ConvertToWav.save_float32_16khz_mono_audio(waveform, voice_file_path)

        return waveform



from abc import abstractmethod
from numpy import ndarray

from ..filter import Filter


class IsolateVoices(Filter):
    """
    Abstract base class for voice isolation methods.

    This class defines the structure for voice isolation methods. Subclasses must implement the 'execute' method.

    Attributes:
        None
    """

    @abstractmethod
    # pylint: disable=arguments-differ
    def execute(self, audio_file_path: str, voice_file_path: str, **kwargs) -> ndarray:
        """
        Execute the voice isolation process.

        Args:
            audio_file_path (str): Path to the source audio file.
            voice_file_path (str): Path to save the isolated voice audio file.
            **kwargs: Additional parameters for customization.

        Returns:
            ndarray: NumPy array representing the isolated voice audio.
        """



from abc import abstractmethod
import numpy as np
from numpy import ndarray
import torch
from torch import Tensor
import soundfile as sf
import torchaudio
from pydub import AudioSegment

from ..filter import Filter


class ConvertToWav(Filter):
    @abstractmethod
    # pylint: disable=arguments-differ
    def execute(self, source_file_path: str, audio_file_path: str, **kwargs: dict) -> None:
        """
        Abstract method for converting audio files to WAV format.

        Args:
            source_file_path (str): Path to the input audio file.
            audio_file_path (str): Path to save the output WAV file.
            **kwargs (dict): Additional parameters for the conversion process.
        """

    @staticmethod
    def format_float32_16khz_mono_audio(audio: AudioSegment, device: str = "cuda") -> ndarray:
        """
        Convert PyDub AudioSegment to a NumPy array with float32 format and 16kHz sample rate.

        Args:
            audio (AudioSegment): Input audio in PyDub's AudioSegment format.

        Returns:
            ndarray: NumPy array representing the audio with float32 format and 16kHz sample rate.
        """
        if device == "cpu":
            np_audio: ndarray = np.frombuffer(audio.set_channels(1).set_sample_width(2).raw_data, dtype=np.int16)
            waveform: Tensor = torch.from_numpy(np_audio)
            waveform = waveform.float() / 32767.0
            resampler = torchaudio.transforms.Resample(orig_freq=audio.frame_rate, new_freq=16000)
            waveform = resampler(waveform)
            return waveform.numpy()
        else:
            np_audio: ndarray = np.frombuffer(audio.set_channels(1).set_sample_width(2).raw_data, dtype=np.int16)
            waveform: Tensor = torch.from_numpy(np_audio).cuda()
            waveform = waveform.float() / 32767.0
            resampler = torchaudio.transforms.Resample(orig_freq=audio.frame_rate, new_freq=16000).cuda()
            waveform = resampler(waveform)
            return waveform.cpu().numpy()

    @staticmethod
    def load_float32_16khz_mono_audio(input_file: str, device:str = "cuda") -> ndarray:
        """
        Load audio from file and convert it to a NumPy array with float32 format and 16kHz sample rate.

        Args:
            input_file (str): Path to the input audio file.

        Returns:
            ndarray: NumPy array representing the audio with float32 format and 16kHz sample rate.
        """
        audio: AudioSegment = AudioSegment.from_file(input_file).set_channels(1).set_sample_width(2)
        return ConvertToWav.format_float32_16khz_mono_audio(audio=audio, device=device)

    @staticmethod
    def save_float32_16khz_mono_audio(waveform: ndarray, wav_file: str) -> None:
        """
        Save a NumPy array with float32 format and 16kHz sample rate to a WAV file.

        Args:
            waveform (ndarray): NumPy array representing the audio with float32 format and 16kHz sample rate.
            wav_file (str): Path to save the output WAV file.
        """
        scaled_data = waveform.astype(np.float32)
        sf.write(wav_file, scaled_data, 16000)



"""
pipeline
"""
#pylint: disable=unused-import
import os
import logging

from .context import Context
from .wav_conversion import ConvertToWav, ConvertToWavFFMpeg, ConvertToWavSoundfile
from .voice_isolation import IsolateVoices, IsolateVoicesNone, IsolateVoicesFile, IsolateVoicesDemucs, IsolateVoicesMDX
from .speaker_diarization import DiarizeSpeakers, DiarizeSpeakersNone, DiarizeSpeakersPyannote
from .speaker_diarization import DiarizeSpeakersFile, DiarizeSpeakersSpeechBrain
from .language_detection import DetectLanguage, DetectLanguageFasterWhisper, DetectLanguageWhisper, DetectLanguageFile
from .speech_transcription import TranscribeSpeechWhisper, TranscribeSpeechFile
from .speech_transcription import TranscribeSpeech, TranscribeSpeechFasterWhisper
from .transcript_writing import WriteTranscript, WriteTranscriptDocx, WriteTranscriptAss, WriteTranscriptStdout
from .transcription import Transcription

LOG = logging.getLogger(__name__)

class Pipeline:
    def __init__(self,
                 context: Context,
                 convert_to_wav: ConvertToWav = None,
                 isolate_voices: IsolateVoices = None,
                 diarize_speakers: DiarizeSpeakers = None,
                 detect_languages: DetectLanguage = None,
                 speech_transcription: TranscribeSpeech = None,
                 transcripte_writing: [WriteTranscript] = None,
                 ):

        if convert_to_wav is None:
            if context.has_ffmpeg:
                convert_to_wav = ConvertToWavFFMpeg()
            else:
                LOG.warning("ffmpeg was not detected, will only handle .wav audio files")
                convert_to_wav = ConvertToWavSoundfile()

        if isolate_voices is None:
            isolate_voices = IsolateVoicesMDX()
        if diarize_speakers is None:
            diarize_speakers = DiarizeSpeakersPyannote()
        if detect_languages is None:
            detect_languages = DetectLanguageFasterWhisper()
        if speech_transcription is None:
            speech_transcription = TranscribeSpeechFasterWhisper()
        if transcripte_writing is None:
            transcripte_writing = [ WriteTranscriptDocx(), WriteTranscriptAss(), WriteTranscriptStdout() ]

        self.context: Context = context
        self.convert_to_wav: ConvertToWav = convert_to_wav
        self.isolate_voices: IsolateVoices = isolate_voices
        self.diarize_speakers: DiarizeSpeakers = diarize_speakers
        self.detect_languages: DetectLanguage = detect_languages
        self.transcript_speech: TranscribeSpeech = speech_transcription
        self.transcripte_writing: [WriteTranscript] = transcripte_writing

    def execute(self):
        os.makedirs(self.context.work_directory_path, exist_ok=True)

        if self.context.transcribe_only:
            self.context.audio_file_path = self.context.source_file_path
            self.context.voice_file_path = self.context.audio_file_path
            self.context.language_file = None
            self.context.diarization_file = None
            filters: [] = [
                self.transcript_speech
            ] + self.transcripte_writing
        else:
            filters: [] = [
                self.convert_to_wav,
                self.isolate_voices,
                self.diarize_speakers,
                self.detect_languages,
                self.transcript_speech
            ] + self.transcripte_writing

        for f in filters:
            f.load(**self.context.to_dict())
            f.execute(**self.context.to_dict())
            f.unload(**self.context.to_dict())



import logging
from typing import List
from docx import Document
from docx.shared import RGBColor
from docx.enum.text import WD_COLOR_INDEX
from docx.oxml import OxmlElement
from langcodes import standardize_tag

from ..transcription import Utterance, Transcription
from .write_transcript import WriteTranscript  # Assuming WriteTranscript is in the same directory.

LOG = logging.getLogger(__name__)

def format_seconds(seconds: float) -> str:
    """
    Format seconds into HH:MM:SS.

    Parameters:
        seconds (float): The time in seconds.

    Returns:
        str: The formatted time.
    """
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return f'{int(hours):02}:{int(minutes):02}:{int(seconds):02}'

def short_code_to_bcp47(short_code: str) -> str:
    """
    Convert a language code from a short format to BCP47.

    Parameters:
        short_code (str): The input language code in a short format.

    Returns:
        str: The BCP47 language code.
    """
    try:
        return standardize_tag(short_code)
    except ValueError as e:
        LOG.error(f"Error converting language code: {e}")
        return short_code  # Returning the original code might be a good fallback.



def format_word(paragraph, word, formatting, lang):
    """
    Format a word in a paragraph based on specified formatting options.

    Parameters:
        paragraph (docx.text.paragraph.Paragraph): The paragraph to add the word to.
        word (str): The word text.
        formatting (dict): Formatting options (color, underline, highlight).
        lang (str): The language code of the word.

    Returns:
        None
    """
    run = paragraph.add_run(word)

    if lang:
        #pylint: disable=protected-access
        lang_element = OxmlElement('w:lang')
        language_code = short_code_to_bcp47(lang)
        lang_element.set('{http://www.w3.org/XML/1998/namespace}lang', language_code)
        run._element.append(lang_element)

    if formatting['color'] is not None:
        run.font.color.rgb = RGBColor(*formatting['color'])

    run.underline = formatting['underline']

    if formatting['highlight'] is not None:
        run.font.highlight_color = formatting['highlight']
    else:
        run.font.highlight_color = WD_COLOR_INDEX.AUTO



def write_docx(utterances: List[Utterance], no_timestamps: bool, no_speakers: bool,
               with_confidence: bool, with_language: bool, output_file: str) -> None:
    """
    Write a list of utterances to a Microsoft Word (docx) file.

    Parameters:
        utterances (List[Utterance]): List of Utterance objects to be written.
        no_timestamps (bool): If True, exclude timestamps from the output.
        no_speakers (bool): If True, exclude speaker information from the output.
        with_confidence (bool): If True, include confidence-related formatting.
        with_language (bool): If True, include language information.
        output_file (str): Path to the output Word file.

    Returns:
        None
    """
    doc = Document()

    for utterance in utterances:
        paragraph = doc.add_paragraph()
        header = ""

        if not no_timestamps:
            header += f"[{format_seconds(utterance.start)}]"
        if not no_speakers:
            header += f" {utterance.speaker}"
        if header:
            header += ":"
            run = paragraph.add_run(header)
            run.bold = True

        for word in utterance.words:
            formatting = {'color': None, 'underline': False, 'highlight': None}

            if with_confidence and word.confidence < 0.80:
                formatting['underline'] = True

            if with_confidence and word.confidence < 0.50:
                formatting['highlight'] = WD_COLOR_INDEX.YELLOW

            lang = utterance.language if with_language else None

            format_word(paragraph, word.text, formatting, lang)

    doc.save(output_file)



class WriteTranscriptDocx(WriteTranscript):
    def __init__(self, no_timestamps=False, no_speakers=False, with_confidence=True, with_language=True):
        """
        Constructor for WriteTranscriptDocx class.

        Parameters:
            no_timestamps (bool): If True, exclude timestamps from the output.
            no_speakers (bool): If True, exclude speaker information from the output.
            with_confidence (bool): If True, include confidence-related formatting.
            with_language (bool): If True, include language information.
        """
        self.no_timestamps = no_timestamps
        self.no_speakers = no_speakers
        self.with_confidence = with_confidence
        self.with_language = with_language

    def execute(self, transcription_path: str, output_file: str, **kwargs: dict):
        """
        Execute the transcription writing process to a Microsoft Word (docx) file.

        Parameters:
            transcript (Transcription): The input transcription to be written.
            output_file (str): Path to the output Word file.
            **kwargs: Additional keyword arguments (not used in this implementation).

        Returns:
            None
        """
        transcript:Transcription = Transcription.load(transcription_path)
        output_file = f"{output_file}.docx"
        write_docx(
            utterances=transcript.utterances,
            no_timestamps=self.no_timestamps,
            no_speakers=self.no_speakers,
            with_confidence=self.with_confidence,
            with_language=self.with_language,
            output_file=output_file
        )



from numpy import ndarray

from ..wav_conversion import ConvertToWav
from .isolate_voices import IsolateVoices


class IsolateVoicesNone(IsolateVoices):
    """
    Passthrough voice isolation.

    This class inherits from IsolateVoices and represents a passthrough implementation where the input audio
    is loaded and saved without any voice isolation.

    Attributes:
        None
    """

    def execute(self, audio_file_path: str, voice_file_path: str, **kwargs) -> ndarray:
        """
        Execute the passthrough voice isolation.

        Args:
            audio_file_path (str): Path to the input audio file.
            voice_file_path (str): Path to save the output audio file.
            **kwargs: Additional parameters (not used in this method).

        Returns:
            ndarray: NumPy array representing the input audio.
        """
        # Load the input audio as a NumPy array
        waveform: ndarray = ConvertToWav.load_float32_16khz_mono_audio(audio_file_path, kwargs['device'])

        # Save the input audio to the destination path
        ConvertToWav.save_float32_16khz_mono_audio(waveform, voice_file_path)

        return waveform



from abc import abstractmethod
from pyannote.core import Annotation

from ..filter  import Filter


class DiarizeSpeakers(Filter):
    """
    Abstract class for diarization of speakers in an audio file.

    This class defines the interface for diarization methods. Subclasses must implement the 'execute' method.

    Attributes:
        None
    """

    @abstractmethod
    # pylint: disable=arguments-differ
    def execute(self, voice_file_path: str, diarization_file: str, min_speakers: int = 1, max_speakers: int = None,
                **kwargs: dict) -> Annotation:
        """
        Execute the diarization process.

        Args:
            voice_file_path (str): Path to the input audio file.
            diarization_file (str): Path to the output RTTM (Rich Transcription Time Marked) file.
            min_speakers (int, optional): Minimum number of expected speakers. Default is 1.
            max_speakers (int, optional): Maximum number of expected speakers. Default is None (unbounded).
            **kwargs (dict): Additional parameters for customization.

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """



"""
__init__.py
"""
from .diarize_speakers import DiarizeSpeakers
from .diarize_speakers_none import DiarizeSpeakersNone
from .diarize_speakers_pyannote import DiarizeSpeakersPyannote
from .diarize_speakers_file import DiarizeSpeakersFile
from .diarize_speakers_speechbrain import DiarizeSpeakersSpeechBrain



import os
import sys
import unittest
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

# pylint: disable=(wrong-import-position
from verbatim.pipeline import Pipeline
from verbatim.context import Context

class TestStringMethods(unittest.TestCase):
    def test_verbatim(self):
        pass

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d][%(funcName)s] %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%SZ')

    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # Set CUDA_VISIBLE_DEVICES to -1 to use CPU

    print(os.getcwd())
    #context: Context = Context(source_file="tests/data/test.mp3", languages=["en", "fr", "es", "it"], nb_speakers=2)
    context: Context = Context(
        source_file="tests/data/init.mp3", languages=["en"], nb_speakers=2, log_level=logging.DEBUG, device="cpu")
    Pipeline(context).execute()



import argparse
import os
import sys
import logging
import torch
from .__init__ import __version__
from .pipeline import Pipeline
from .context import Context


LOG = logging.getLogger(__name__)

# Get the package name dynamically
package_name = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

def main():
    parser = argparse.ArgumentParser(prog=package_name, description="Verbatim: that's what she said")

    # Specify the command line arguments
    parser.add_argument("input",
                        help="Path to the input audio file")
    parser.add_argument("-o", "--output",
                        help="Path to the output directory", default=".")
    parser.add_argument("-l", "--languages", nargs="*",
                        help="Languages for speech recognition. Provide multiple values for multiple languages.")
    parser.add_argument("-n", "--nb_speakers", type=int,
                        help="Number of speakers in the audio file. Defaults to 1.", default=1)
    parser.add_argument("-b", "--nb_beams", type=int,
                        help="Number of parallel when resolving transcription. " +
                        "1-3 for fast, 12-15 for high quality. Default is 9.",
                        default=None)
    parser.add_argument("-v", "--verbose", action="count", default=0,
                        help="Increase verbosity (specify multiple times for more verbosity)")
    parser.add_argument("--version", action="version", version=f"{package_name} {__version__}")
    parser.add_argument("--cpu", action="store_true", help="Toggle CPU usage")
    parser.add_argument("--transcribe_only", action="store_true",
                        help="Skip preprocessing, including diarization and only transcribe")
    parser.add_argument("--write-config",
                        help="Write the config file so it can be edited", default=None)
    parser.add_argument("--read-config",
                        help="Read a config file", default=None)

    args = parser.parse_args()

    # Set logging level based on verbosity
    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = log_levels[min(args.verbose, len(log_levels) - 1)]
    logging.basicConfig(stream=sys.stderr,
                        level=log_level,
                        format='%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d][%(funcName)s] %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%SZ')

    # Check if the version option is specified
    if hasattr(args, 'version') and args.version:
        return  # Exit if the version option is specified

    # Print help if no input is provided
    if not args.input:
        parser.print_help()
        return

    # Validate input file existence
    if not os.path.isfile(args.input):
        print(f"Error: Input file '{args.input}' not found.")
        return

    # Validate output directory existence
    if not os.path.isdir(args.output):
        print(f"Error: Output directory '{args.output}' not found.")
        return

    # Create output directory if it does not exist
    if not os.path.exists(args.output):
        os.makedirs(args.output)

    if args.cpu or not torch.cuda.is_available():
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # Set CUDA_VISIBLE_DEVICES to -1 to use CPU
        print("Using CPU")
        device = "cpu"
    else:
        print("Using GPU")
        device = "cuda"

    # Use the provided values or default values for languages and nb_speakers
    languages = args.languages if args.languages else ["en", "fr"]
    nb_speakers = args.nb_speakers

    if args.read_config:
        with open(args.read_config, "r", encoding="utf-8") as f:
            text = f.read()
            context: Context = Context.from_yaml(source_file=args.input, out_dir=args.output, text=text)
    else:
        context: Context = Context(source_file=args.input, out_dir=args.output)

    context.languages=languages
    context.nb_speakers=nb_speakers
    context.log_level=log_level
    context.device=device

    if args.nb_beams is None:
        if context.beams is None:
            context.beams = 9
    else:
        context.beams = args.nb_beams

    if args.transcribe_only:
        context.transcribe_only = True
    else:
        context.transcribe_only = False


    if args.write_config:
        with open(args.write_config, "w", encoding="utf-8") as f:
            f.write(context.to_yaml())

    pipeline: Pipeline = Pipeline(context=context)
    pipeline.execute()

if __name__ == "__main__":
    main()



import sys
import os
import logging
import torchaudio
import numpy as np

from speechbrain.pretrained import VAD
from pyannote.core import Annotation, Segment

from ..wav_conversion import ConvertToWav
from .diarize_speakers import DiarizeSpeakers


LOG = logging.getLogger(__name__)


class DiarizeSpeakersSpeechBrain(DiarizeSpeakers):
    """
    Diarization implementation using SpeechBrain for speaker segmentation.

    This class inherits from DiarizeSpeakers and implements diarization using SpeechBrain for speaker segmentation.

    Attributes:
        None
    """

    def diarize_on_silences(self, audio_path: str, model_speechbrain_vad:str, **kwargs:dict) -> Annotation:
        """
        Diarize speakers based on silences using SpeechBrain.

        Args:
            audio_file (str): Path to the input audio file.

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        # Set up temporary directory for VAD model
        tmpdir = os.path.join(kwargs['work_directory_path'], "tmpdir")

        # Load VAD model from SpeechBrain
        try:
            vad_model = VAD.from_hparams(
                source=model_speechbrain_vad,
                savedir=tmpdir,
                run_opts={"device": kwargs['device']}
            )
        except OSError as e:
            if sys.platform == "win32":
                # pylint: disable=no-member
                if e.winerror == 1314:
                    LOG.exception( "Not enough priviledges to install the model. " +
                                  f"Audio will not be segmented on silences: {str(e)}")
                    waveform = ConvertToWav.load_float32_16khz_mono_audio(
                        input_file=audio_path,
                        device=kwargs['device'])
                    diarization = Annotation("waveform")
                    diarization[Segment(0, len(waveform)/16000)] = "speaker"
                    return diarization
            raise

        # Perform VAD
        boundaries = vad_model.get_speech_segments(audio_path)

        # Create a Pyannote Annotation object for diarization
        diarization = Annotation()

        # Add speaker segments to the diarization annotation
        for i in range(0, len(boundaries), 2):
            if i+1 < len(boundaries):
                diarization[Segment(float(boundaries[i][0]), float(boundaries[i + 1][1]))] = "speaker"
            else:
                diarization[Segment(float(boundaries[i][0]), float(boundaries.data[0][1]))] = "speaker"

        # Upsample boundaries and save the VAD result as a new audio file
        upsampled_boundaries = vad_model.upsample_boundaries(boundaries=boundaries, audio_file=audio_path)
        torchaudio.save(f"{audio_path}_vad.wav", upsampled_boundaries.cpu(), 16000)

        return diarization

    @staticmethod
    def pad_audio_to_duration(audio_samples, target_duration_seconds, sampling_rate=16000):
        # Calculate the target length in samples
        target_length_samples = int(target_duration_seconds * sampling_rate)

        # Calculate the current length of the audio
        current_length_samples = len(audio_samples)

        # Calculate the amount of padding needed
        padding_needed = target_length_samples - current_length_samples
        if padding_needed <= 0:
            return audio_samples

        # Pad the audio samples with zeros (silence)
        padded_audio = np.pad(audio_samples, (0, padding_needed), mode='constant', constant_values=0)

        return padded_audio

    def execute(self, voice_file_path: str, diarization_file: str,
                min_speakers: int = 1, max_speakers: int = None,
                model_speechbrain_vad:str  = None, **kwargs: dict) -> Annotation:
        """
        Execute the diarization process using SpeechBrain.

        Args:
            voice_file_path (str): Path to the input audio file.
            diarization_file (str): Path to the output RTTM (Rich Transcription Time Marked) file.
            min_speakers (int, optional): Minimum number of expected speakers. Default is 1.
            max_speakers (int, optional): Maximum number of expected speakers. Default is None (unbounded).
            **kwargs (dict): Additional parameters (not used in this method).

        Returns:
            Annotation: Pyannote Annotation object containing information about speaker diarization.
        """
        # Perform diarization based on silences using SpeechBrain
        diarization: Annotation = self.diarize_on_silences(audio_path=voice_file_path,
                                                           model_speechbrain_vad=model_speechbrain_vad)

        # Write the diarization result to the output RTTM file
        with open(diarization_file, "w", encoding="utf-8") as f:
            diarization.write_rttm(f)

        # Log the diarization result for information
        LOG.info(diarization)

        return diarization



import sys
from ..transcription import Transcription
from .write_transcript import WriteTranscript  # Assuming WriteTranscript is in the same directory.

class WriteTranscriptStdout(WriteTranscript):
    def __init__(self, with_colours=True):
        self.with_colours = with_colours

    def execute(self, transcription_path: str, output_file: str, **kwargs: dict):
        transcript:Transcription = Transcription.load(transcription_path)
        if self.with_colours:
            text:str = transcript.get_colour_text()
        else:
            text:str = transcript.get_text()

        sys.stdout.write(text)
        sys.stdout.write("\n")
        sys.stdout.flush()



"""
setuptools module for uniscripts.
"""
from pathlib import Path
import re
from setuptools import setup, find_packages

PACKAGE_NAME = "verbatim"
INIT_FILE = Path(f"{PACKAGE_NAME}/__init__.py").absolute()
REQUIREMENTS = """
    wheel
    onnxruntime-gpu==1.16.0
    onnxruntime
    numpy

    pyannote.core
    pyannote.audio

    torch
    torchaudio

    faster_whisper
    openai-whisper

    soundfile
    demucs

    dataclasses_json

    python-ffmpeg
    pydub

    audio_separator

    python-docx
    langcodes
""".split()


# Load version from __init__.py
print(INIT_FILE)
with open(INIT_FILE, encoding="utf-8") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        f.read(), re.MULTILINE).group(1)

# Load load description from README.md
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")



setup(
    name=PACKAGE_NAME,
    version=version,
    description='high quality multi-lingual speech to text',
    entry_points={
            'console_scripts': [
                'verbatim = verbatim.__main__:main',
            ],
        },
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=REQUIREMENTS,

    url='https://github.com/gaspardpetit/verbatim',

    # Author details
    author='Gaspard Petit',
    author_email='gaspardpetit@gmail.com',

    # Choose your license
    license='MIT',

    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        "Intended Audience :: Developers",
        'Topic :: Multimedia :: Sound/Audio :: Speech',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',

        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        "Operating System :: OS Independent",

        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",

    ],

    keywords=[
        'speech-to-text',
        'audio processing',
        'multilingual',
        'natural language processing',
        'automatic speech recognition',
        'ASR',
        'text transcription',
        'language support',
        'machine learning',
        'deep learning',
        'NLP',
        'linguistics',
        'voice recognition',
        'PyTorch',
        'TensorFlow',
        'audio analysis',
        'speech analytics',
        'spoken language processing',
        'i18n',
        'internationalization',
    ],

)



import logging
import torch
from numpy import ndarray
from whisper.audio import pad_or_trim, N_FRAMES, N_SAMPLES, log_mel_spectrogram

from .detect_language import DetectLanguage
from ..models.model_whisper import WhisperModel
from ..transcription import Transcription, Utterance


LOG = logging.getLogger(__name__)

class DetectLanguageWhisper(DetectLanguage):
    """
    Class for language detection using the Whisper model.

    Attributes:
        None
    """

    def execute_segment(self, speaker: str, speech_offset: float, speech_segment_float32_16khz: ndarray,
                        languages=None, **kwargs: dict) -> Transcription:
        """
        Executes language detection on a speech segment using the Whisper model.

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language_file (str): File path to load the detected language information.
            languages (Optional[List[str]]): List of target languages for detection.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription object containing the detected language information.
        """
        transcription = Transcription()

        # Load Whisper model
        if 'device' in kwargs:
            WhisperModel.device = kwargs['device']
        if 'model_whisper' in kwargs:
            WhisperModel.model = kwargs['model_whisper']
        model = WhisperModel().model

        if kwargs['device'] == "cpu":
            float_type: torch.dtype = torch.float32
        else:
            float_type: torch.dtype = torch.float16

        # Generate mel spectrogram from the speech segment
        mel = log_mel_spectrogram(speech_segment_float32_16khz, model.dims.n_mels, padding=N_SAMPLES)
        mel_segment = pad_or_trim(mel, N_FRAMES).to(model.device).to(float_type)

        # Perform language detection using the Whisper model
        whisper_transcription = model.detect_language(mel=mel_segment)
        detected_languages = whisper_transcription[1]

        # Determine the best language based on user-specified languages or the maximum probability
        best_language, best_language_prob = self.determine_best_language(detected_languages, languages)

        # Create an Utterance object with the detected language information
        utterance = Utterance(
            start=speech_offset,
            end=speech_offset + len(speech_segment_float32_16khz) / 16000,
            language=best_language,
            confidence=best_language_prob,
            speaker=speaker,
            words=[])

        # Append the Utterance to the transcription
        transcription.append(utterance)

        # Log information about the detected utterance
        LOG.info(utterance.get_colour_text())

        return transcription

    def determine_best_language(self, detected_languages, target_languages):
        """
        Determines the best language based on detected languages and user-specified target languages.

        Args:
            detected_languages (dict): Dictionary of detected languages and their probabilities.
            target_languages (List[str]): List of target languages.

        Returns:
            Tuple[str, float]: The best language and its probability.
        """
        best_language = None
        best_language_prob = 0

        if target_languages and len(target_languages) > 0:
            for lang in target_languages:
                if lang in detected_languages:
                    if detected_languages[lang] > best_language_prob:
                        best_language_prob = detected_languages[lang]
                        best_language = lang

        if best_language is None:
            best_language = max(detected_languages, key=detected_languages.get)
            best_language_prob = detected_languages[best_language]

        return best_language, best_language_prob



from abc import abstractmethod
from ..filter import Filter

class WriteTranscript(Filter):
    """
    Abstract base class for writing transcriptions to a file.

    Attributes:
        None

    Methods:
        execute(transcript: Transcription, output_file: str, **kwargs: dict) -> None:
            Abstract method for executing the transcription writing process.

    """

    @abstractmethod
    # pylint: disable=arguments-differ
    def execute(self, transcription_path: str, output_file: str, **kwargs: dict) -> None:
        """
        Execute the transcription writing process.

        Args:
            transcript (Transcription): The transcription to be written.
            output_file (str): The path to the output file.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            None

        Raises:
            NotImplementedError: This method must be implemented by the derived class.
        """



import logging
from numpy import ndarray

from .transcribe_speech import TranscribeSpeech
from ..transcription import Transcription

LOG = logging.getLogger(__name__)

class TranscribeSpeechFile(TranscribeSpeech):
    """
    Implementation of TranscribeSpeech that loads transcriptions from a file.

    Attributes:
        None
    """

    def execute_segment(self, speech_segment_float32_16khz: ndarray,
                        speaker: str = "speaker", speech_offset: float = 0,
                        language: str = None, prompt: str = "",
                        **kwargs: dict) -> Transcription:
        """
        Execute transcription on a speech segment (Not implemented).

        Args:
            speaker (str): The speaker identifier.
            speech_offset (float): The offset in seconds from the beginning of the audio.
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            language (str): Target language for transcription.
            prompt (str): Optional transcription prompt.
            **kwargs (dict): Additional keyword arguments for customization.

        Raises:
            NotImplementedError: This method is not implemented for loading from a file.

        Returns:
            None
        """
        raise NotImplementedError("execute_segment method is not implemented for loading from a file.")

    # pylint: disable=arguments-differ
    def execute(self, transcription_path: str, **kwargs: dict) -> Transcription:
        """
        Load transcriptions from a file.

        Args:
            speech_segment_float32_16khz (ndarray): Speech segment data in float32 format at 16kHz.
            detected_languages (Transcription): Transcription containing detected languages.
            transcription_path (str): File path for loading transcriptions.
            diarization (Annotation): Diarization information.
            languages (list): List of target languages.
            **kwargs (dict): Additional keyword arguments for customization.

        Returns:
            Transcription: Transcription loaded from the specified file.
        """
        transcription = Transcription.load(transcription_path)
        return transcription



import os
import logging
from ffmpeg import FFmpeg, Progress
from ..wav_conversion.convert_to_wav import ConvertToWav

LOG = logging.getLogger(__name__)

class FFMpegInstance:
    """
    Singleton class to manage a single instance of FFmpeg.

    This class ensures that only one instance of FFmpeg is created and used throughout the application.
    """

    __instance = None

    def __new__(cls):
        """
        Create a new instance of FFMpegInstance if one does not exist, otherwise return the existing instance.

        Returns:
            FFMpegInstance: The singleton instance of FFMpegInstance.
        """
        if cls.__instance is None:
            cls.__instance = super(FFMpegInstance, cls).__new__(cls)
            cls.__instance._init_once()
        return cls.__instance

    # pylint: disable=attribute-defined-outside-init
    def _init_once(self):
        """
        Initialize the singleton instance with an instance of FFmpeg and set up event listeners.
        """
        self.ffmpeg: FFmpeg = FFmpeg().option("y")
        self.ffmpeg.add_listener("progress", FFMpegInstance._on_progress)
        self.ffmpeg.add_listener("start", FFMpegInstance._on_start)
        self.ffmpeg.add_listener("completed", FFMpegInstance._on_completed)
        self.ffmpeg.add_listener("terminated", FFMpegInstance._on_terminated)
        self.ffmpeg.add_listener("stderr", FFMpegInstance._on_stderr)

    @staticmethod
    def _on_terminated():
        """
        Callback function called when ffmpeg is terminated.
        """
        LOG.warning("ffmpeg was terminated")

    @staticmethod
    def _on_completed():
        """
        Callback function called when ffmpeg completes successfully.
        """
        LOG.info("ffmpeg completed")

    @staticmethod
    def _on_progress(progress: Progress):
        """
        Callback function called to report the progress of ffmpeg.

        Args:
            progress (Progress): Progress object containing information about the conversion progress.
        """
        LOG.info(progress)

    @staticmethod
    def _on_start(arguments: list[str]):
        """
        Callback function called when ffmpeg starts.

        Args:
            arguments (list[str]): List of arguments used to start ffmpeg.
        """
        LOG.info(f"ffmpeg started: {' '.join(arguments)}")

    @staticmethod
    def _on_stderr(text: str):
        """
        Callback function called when there is an error message in ffmpeg's stderr.

        Args:
            text (str): Error message from ffmpeg's stderr.
        """
        LOG.debug(text)


class ConvertToWavFFMpeg(ConvertToWav):

    def execute(self, source_file_path: str, audio_file_path: str, **kwargs: dict):
        """
        Convert audio file to WAV format using ffmpeg.

        Args:
            source_file_path (str): Path to the input audio file.
            audio_file_path (str): Path to save the output WAV file.
            **kwargs (dict): Additional parameters (not used in this method).
        """
        LOG.info(f"Converting {source_file_path} to {audio_file_path}")

        # Use ffmpeg from the singleton instance to convert input file to raw PCM
        output_directory = os.path.dirname(audio_file_path)
        os.makedirs(output_directory, exist_ok=True)

        ffmpeg = FFMpegInstance().ffmpeg
        ffmpeg_command = ffmpeg.input(source_file_path).output(audio_file_path, ac='1', acodec='pcm_s32le')
        ffmpeg_command.execute()



import os
import logging
import soundfile as sf
import numpy as np

from ..wav_conversion.convert_to_wav import ConvertToWav

LOG = logging.getLogger(__name__)

class ConvertToWavSoundfile(ConvertToWav):

    def execute(self, source_file_path: str, audio_file_path: str, **kwargs: dict):
        # Use ffmpeg from the singleton instance to convert input file to raw PCM
        output_directory = os.path.dirname(audio_file_path)
        os.makedirs(output_directory, exist_ok=True)
        data, samplerate = sf.read(source_file_path)

        if len(data.shape) > 1:
            data = np.mean(data, axis=1)

        if data.dtype.kind == 'f':
            scaled_data = np.int32(data * np.iinfo(np.int32).max)
        elif data.dtype.kind == 'i':
            if data.dtype.itemsize == 4:
                scaled_data = data
            else:
                scaled_data = np.int32(data.astype(np.float32) / (2 ** (8 * data.dtype.itemsize - 1)))
        else:
            raise ValueError("Unsupported data type")

        sf.write(audio_file_path, scaled_data, samplerate, subtype='PCM_32')



import whisper

class WhisperModel:
    """
    A singleton class for managing an instance of the OpenAI Whisper model.

    Attributes:
        model (whisper.Whisper): The OpenAI Whisper model instance.
    """
    _instance = None
    device: str = "cuda"
    model: str = "large"

    def __new__(cls):
        """
        Create a new instance of the class if it doesn't exist, otherwise, return the existing instance.

        Returns:
            WhisperModel: The WhisperModel instance.
        """
        if not cls._instance:
            cls._instance = super(WhisperModel, cls).__new__(cls)
            cls._instance._init_once(device=WhisperModel.device, model=WhisperModel.model)
        return cls._instance

    # pylint: disable=attribute-defined-outside-init
    def _init_once(self, model:str, device:str):
        """
        Initialize the WhisperModel instance.
        """
        self.model = whisper.load_model(name=model, device=device)

    def unload(self):
        """
        Unload the model by setting the model attribute to None and clearing the instance.
        """
        self.model = None
        WhisperModel._instance = None



from .transcribe_speech import TranscribeSpeech
from .transcribe_speech_fasterwhisper import TranscribeSpeechFasterWhisper
from .transcribe_speech_whisper import TranscribeSpeechWhisper
from .transcribe_speech_file import TranscribeSpeechFile
